library(ggplot2)

dates <- data.frame(date=seq(as.Date('2020-01-01'),as.Date('2020-04-30'),by=1))
dates$day <- factor(strftime(dates$date,format="%a"),levels=rev(c("Mon","Tue","Wed","Thu","Fri","Sat","Sun")))
dates$week <- factor(strftime(dates$date,format="%V"))
dates$month <- factor(strftime(dates$date,format="%B"),levels=c("January","February","March","April"))
dates$ddate <- factor(strftime(dates$date,format="%d"))

dates$comment <- "Available"
dates$comment[dates$date>=as.Date('2020-01-11') & dates$date<=as.Date('2020-01-20')] <- "Limited"
dates$comment[dates$date>=as.Date('2020-03-21') & dates$date<=as.Date('2020-03-29')] <- "Not available"
dates$comment[dates$date>=as.Date('2020-01-01') & dates$date<=as.Date('2020-01-08')] <- "Not available"
dates$comment[dates$day=="Sat" | dates$day=="Sun"] <- "Weekend"

dates$comment <- factor(dates$comment,levels=c("Weekend","Available","Limited","Not available"))

p <- ggplot(dates,aes(x=week,y=day))+
      geom_tile(aes(fill=comment))+
      geom_text(aes(label=ddate))+
      scale_fill_manual(values=c("#86d3c7","#ddffb3","#fb8072","#a3a3b3"))+
      facet_grid(~month,scales="free",space="free")+
      labs(x="Week",y="")+
      theme_bw(base_size=10)+
      theme(legend.title=element_blank(),
            strip.background=element_blank(),
            legend.position="bottom",
            legend.justification="left",
            legend.direction="horizontal",
            legend.key.size=unit(0.15,"cm"),
            legend.spacing.x=unit(0.15,"cm"))

print(p)
